#include "libs.h"
#include "body.h"
#include "frame.h"
#include "serializer.h"
#include "star.h"
#include "planet.h"
#include "space_station.h"
#include "ship.h"
#include "player.h"

Body::Body(void) {
  m_frame = 0;
  m_flags = 0;
  m_projectedPos = vector3d(0.0f, 0.0f, 0.0f);
  m_onscreen = false;
  m_dead = false;
}

Body::~Body(void) {
  /* Do not call delete body. Call Space::KillBody(body). */
  assert(m_dead);
}

void Body::Save(void) {
  using namespace Serializer::Write;
  wr_int(Serializer::LookupFrame(m_frame));
  wr_string(m_label);
  wr_bool(m_onscreen);
  wr_vector3d(m_projectedPos);
  wr_bool(m_dead);
}

void Body::Load(void) {
  using namespace Serializer::Read;
  m_frame         = Serializer::LookupFrame(rd_int());
  m_label         = rd_string();
  m_onscreen      = rd_bool();
  m_projectedPos  = rd_vector3d();
  m_dead          = rd_bool();
}

void Body::Serialize(void) {
  using namespace Serializer::Write;
  wr_int((int)GetType());
  switch(GetType()) {
    case Object::STAR:
    case Object::PLANET:
    case Object::SPACESTATION:
    case Object::SHIP:
    case Object::PLAYER:
      Save();
      break;
    default:
      assert(0);
  }
  wr_vector3d(GetPosition());
  matrix4x4d m;
  GetRotMatrix(m);
  for(int i = 0; i < 16; i++) wr_double(m[i]);
}

Body* Body::Unserialize(void) {
  Body* b = 0;
  using namespace Serializer::Read;
  Object::Type type = (Object::Type)rd_int();
  switch(type) {
    case Object::STAR:
      b = new Star(); break;
    case Object::PLANET:
      b = new Planet(); break;
    case Object::SPACESTATION:
      b = new SpaceStation(); break;
    case Object::SHIP:
      b = new Ship(); break;
    case Object::PLAYER:
      b = new Player(); break;
    default:
      /* TODO: should assert. */
      return 0;
  }
  b->Load();
  /* Must SetFrame() correctly so ModelBodies can add geom to space. */
  Frame* f = b->m_frame;
  b->m_frame = 0;
  b->SetFrame(f);
  b->SetPosition(rd_vector3d());
  matrix4x4d m;
  for(int i = 0; i < 16; i++) m[i] = rd_double();
  b->SetRotMatrix(m);
  return b;
}

/* f == NULL, then absolute position within system. */
vector3d Body::GetPositionRelTo(const Frame* relTo) {
  matrix4x4d m;
  Frame::GetFrameTransform(m_frame, relTo, m);
  return m * GetPosition();
}

const vector3d& Body::GetProjectedPos(void) const {
  assert(IsOnscreen());
  return m_projectedPos;
}

void Body::OrientOnSurface(double radius, double latitude, double longitude) {
  vector3d pos = vector3d(radius*cos(latitude)*cos(longitude),
      radius*sin(latitude)*cos(longitude), radius*sin(longitude));
  vector3d up = vector3d::Normalize(pos);
  SetPosition(pos);

  vector3d forward = vector3d(0,0,1);
  vector3d other = vector3d::Normalize(vector3d::Cross(up, forward));
  forward = vector3d::Cross(other, up);

  matrix4x4d rot = matrix4x4d::MakeRotMatrix(other, up, forward);
  rot = rot.InverseOf();
  SetRotMatrix(rot);
}

